'use strict';

/* For ESLint: List any global identifiers used in this file below */
/* global browser, log, Filter, Fennec, determineUserLanguage, getUnicodeDomain,
FilterNormalizer, STATS, SelectorFilter, setDefault, handlerBehaviorChanged,
SnippetFilter,
BlockingFilterSet, FilterSet, getCustomFiltersText, idleHandler, getSettings */

// Requires jquery and must be on a page with access to the background page

// MyFilters class manages subscriptions and the FilterSet.

// Constructor: merge the stored subscription data and the total possible
// list of subscriptions into this.subscriptions.  Store to disk.
// Inputs: none.
const HOUR_IN_MS = 1000 * 60 * 60;
const keyPrefix = 'filter_list_';
const keyPrefixLength = keyPrefix.length;
const whiteSpaceRE = /\s/;

function MyFilters(callback) {
  const localThis = this;
  this.promise = new Promise(((resolve) => {
    localThis.readyComplete = resolve;
  }));

  this.getSubscriptions((subs) => {
    localThis.subscriptions = subs;

    localThis.officialOptions = localThis.makeSubscriptionOptions();
    if (typeof callback === 'function') {
      callback();
    }
  });
}

// Update _subscriptions and _official_options in case there are changes.
// Should be invoked right after creating a MyFilters object.
MyFilters.prototype.init = function myFiltersInit(callback) {
  this.updateDefaultSubscriptions();
  this.updateFieldsFromOriginalOptions();

  const localThis = this;
  // Build the filter list
  this.onSubscriptionChange(true);

  // On startup and then every hour, check if a list is out of date and has to
  // be updated
  if (this.newUser) {
    this.checkFilterUpdates();
  } else {
    idleHandler.scheduleItemOnce(
      () => {
        localThis.checkFilterUpdates();
      },

      60,
    );
  }

  window.setInterval(
    () => {
      idleHandler.scheduleItemOnce(() => {
        localThis.checkFilterUpdates();
      });
    },

    60 * 60 * 1000,
  );

  if (typeof callback === 'function') {
    callback();
  }
};

// Update the url and requiresList for entries in _subscriptions using values
// from _official_options.
MyFilters.prototype.updateFieldsFromOriginalOptions = function myFitlersUpdateFromOriginal() {
  // Use the stored properties, and only add any new properties and/or lists
  // if they didn't exist in this.subscriptions
  for (const id in this.officialOptions) {
    if (!this.subscriptions[id]) {
      this.subscriptions[id] = {};
    }
    const sub = this.subscriptions[id];
    const official = this.officialOptions[id];

    sub.initialUrl = sub.initialUrl || official.url;
    sub.url = sub.url || official.url;
    if (sub.initialUrl !== official.url) {
      // The official URL was changed. Use it. In case of a redirect, this
      // doesn't happen as only sub.url is changed, not sub.initialUrl.
      sub.initialUrl = official.url;
      sub.url = official.url;
    }

    const isMissingRequiredList = (sub.requiresList !== official.requiresList);
    if (official.requiresList && isMissingRequiredList && sub.subscribed) {
      // A required list was added.  Make sure main list subscribers get it.
      if (this.subscriptions[official.requiresList]) {
        this.changeSubscription(official.requiresList, { subscribed: true });
      }
    }

    sub.requiresList = official.requiresList;
    sub.subscribed = sub.subscribed || false;
    sub.allowSnippets = official.allowSnippets || false;
  }
};

// Update default subscriptions in the browser storage.
// Removes subscriptions localThis are no longer in the official list, not user
// submitted and no longer subscribed.
// Also, converts user submitted subscriptions to recognized one if it is already
// added to the official list and vice-versa.
MyFilters.prototype.updateDefaultSubscriptions = function myFiltersUpdateDefaultSubs() {
  if (this.newUser) {
    // Brand new user. Install some filters for them.
    this.subscriptions = this.loadDefaultSubscriptions();
    return;
  }

  for (const id in this.subscriptions) {
    // Delete unsubscribed ex-official lists.
    if (!this.officialOptions[id] && !this.subscriptions[id].userSubmitted
        && !this.subscriptions[id].subscribed) {
      delete this.subscriptions[id];
    } else {
      // Convert subscribed ex-official lists into user-submitted lists.
      // Convert subscribed ex-user-submitted lists into official lists.
      // Cache subscription that needs to be checked.
      const subToCheck = this.subscriptions[id];
      let isUserSubmitted = true;
      let updateID = id;
      if (!this.officialOptions[id]) {
        // If id is not in official options, check if there's a matching url in the
        // official list. If there is, then the subscription is not user submitted.
        for (const officialID in this.officialOptions) {
          const officialURL = this.officialOptions[officialID].url;
          if (subToCheck.initialUrl === officialURL
            || subToCheck.url === officialURL) {
            isUserSubmitted = false;
            updateID = officialID;
            break;
          }
        }
      } else {
        isUserSubmitted = false;
      }

      subToCheck.userSubmitted = isUserSubmitted;

      // Function that will add a new entry with updated id,
      // and will remove old entry with outdated id.
      const localThis = this;
      const renameSubscription = function (oldId, newId) {
        localThis.subscriptions[newId] = localThis.subscriptions[oldId];
        delete localThis.subscriptions[oldId];
      };

      // Create new id and check if new id is the same as id.
      // If not, update entry in subscriptions.
      const newID = isUserSubmitted ? (`url:${subToCheck.url}`) : updateID;

      if (newID !== id) {
        renameSubscription(id, newID);
      }
    }
  }
};

MyFilters.prototype.ready = function myFiltersReady() {
  return this.promise;
};
// When a subscription property changes, this function stores it
// Inputs: rebuild? boolean, true if the filterset should be rebuilt
MyFilters.prototype.onSubscriptionChange = function myFiltersSubChange(rebuild) {
  this.saveSubscriptions();

  // The only reasons to (re)build the filter set are
  // - when AdBlock starts
  // - when a filter list text is changed ([un]subscribed or updated a list)
  if (rebuild) {
    this.rebuild();
  }

  browser.runtime.sendMessage({ command: 'filters_updated' }).catch((err) => {
    log(`filters_updated listener not yet ready: ${err}`);
  });
};

// get filters that are defined in the extension
MyFilters.prototype.getExtensionFilters = function myFiltersExtensionFilters(settings) {
  const texts = [];
  if (settings.whitelist_hulu_ads) {
    // Issue 7178: FilterNormalizer removes EasyList's too-broad Hulu whitelist
    // entries.  If the user enables whitelist_hulu_ads, just add them back.
    // This workaround can be removed when EasyList changes its Hulu strategy.
    texts.push('@@||ads.hulu.com/published/*.flv');
    texts.push('@@||ads.hulu.com/published/*.mp4');
    texts.push('@@||ll.a.hulu.com/published/*.flv');
    texts.push('@@||ll.a.hulu.com/published/*.mp4');
  }

  return texts;
};

// Rebuild filters based on the current settings and subscriptions.
MyFilters.prototype.rebuild = function myFiltersRebuild() {
  let texts = [];

  // Only add subscriptions in Chrome, Opera, and older version of Safari...
  for (const id in this.subscriptions) {
    if (this.subscriptions[id].subscribed) {
      texts.push(this.subscriptions[id].text);
    }
  }

  // Include custom filters.
  const localThis = this;
  getCustomFiltersText((customfilters) => {
    if (customfilters) {
      texts.push(FilterNormalizer.normalizeList(customfilters, false, true)); // allow snippets
    }

    texts = texts.concat(localThis.getExtensionFilters(getSettings()));
    texts = texts.join('\n').split('\n');

    const filters = localThis.splitByType(texts);

    localThis.hiding = FilterSet.fromFilters(filters.hiding);

    localThis.advanceHiding = FilterSet.fromFilters(filters.advanceHiding);

    localThis.snippets = FilterSet.fromFilters(filters.snippets);

    localThis.blocking = new BlockingFilterSet(
      FilterSet.fromFilters(filters.pattern),
      FilterSet.fromFilters(filters.whitelist),
    );

    handlerBehaviorChanged(); // defined in background

    localThis.readyComplete();

    // if the user is subscribed to malware, then get it
    if (localThis.subscriptions
        && localThis.subscriptions.malware
        && localThis.subscriptions.malware.subscribed
        && !localThis.getMalwareDomains()) {
      localThis.initializeMalwareDomains();
    }

    // After 90 seconds, delete the cache. That way the cache is available when
    // rebuilding multiple times in a row (when multiple lists have to update at
    // the same time), but we save memory during all other times.
    window.setTimeout(() => {
      Filter.cache = {};
    }, 90000);
  });
};

MyFilters.prototype.splitByType = function myFiltersSplitType(texts) {
  // Remove duplicates and empties.
  const unique = {};
  for (let i = 0; i < texts.length; i += 1) {
    unique[texts[i]] = 1;
  }

  delete unique[''];
  const hidingUnmerged = [];
  const advanceHidingUnmerged = [];
  const snippetsUnmerged = [];

  const filters = {
    hiding: {},
    exclude: {},
    advanceHiding: {},
    pattern: {},
    snippets: {},
    whitelist: {},
  };
  let filter;
  for (const text in unique) {
    filter = Filter.fromText(text);
    if (Filter.isSelectorExcludeFilter(text)) {
      setDefault(filters.exclude, filter.selector, []).push(filter);
    } else if (Filter.isSelectorFilter(text)) {
      hidingUnmerged.push(filter);
    } else if (Filter.isAdvancedSelectorFilter(text)) {
      advanceHidingUnmerged.push(filter);
    } else if (Filter.isContentFilter(text)) {
      snippetsUnmerged.push(filter);
    } else if (Filter.isWhitelistFilter(text)) {
      filters.whitelist[filter.id] = filter;
    } else {
      filters.pattern[filter.id] = filter;
    }
  }

  for (let i = 0; i < hidingUnmerged.length; i += 1) {
    filter = hidingUnmerged[i];
    const hider = SelectorFilter.merge(filter, filters.exclude[filter.selector]);
    filters.hiding[hider.id] = hider;
  }
  for (let i = 0; i < advanceHidingUnmerged.length; i += 1) {
    filter = advanceHidingUnmerged[i];
    const hider = SelectorFilter.merge(filter, filters.exclude[filter.selector]);
    filters.advanceHiding[hider.id] = hider;
  }
  for (let i = 0; i < snippetsUnmerged.length; i += 1) {
    filter = snippetsUnmerged[i];
    const snippet = SnippetFilter.merge(filter, filters.exclude[filter.selector]);
    filters.snippets[snippet.id] = snippet;
  }
  return filters;
};

// Change a property of a subscription or check if it has to be updated
// Inputs: id: the id of the subscription to change
//         subData: object containing all data that should be changed
//         forceFetch: if the subscriptions have to be fetched again forced
MyFilters.prototype.changeSubscription = function myFiltersChangeSub(id, subData, forceFetch) {
  let subscribeRequiredListToo = false;
  let listDidntExistBefore = false;

  // Check if the list has to be updated
  function outOfDate(subscription) {
    if (forceFetch) {
      return true;
    }

    // After a failure, wait at least a day to refetch (overridden below if
    // it's a new filter list, having no .text)
    const failedAt = subscription.lastUpdateFailedAt || 0;
    if (Date.now() - failedAt < HOUR_IN_MS * 24) {
      return false;
    }

    // Don't let expiresAfterHours delay indefinitely (Issue 7443)
    const hardStop = subscription.expiresAfterHoursHard || 240;
    const smallerExpiry = Math.min(subscription.expiresAfterHours, hardStop);
    const millis = Date.now() - subscription.lastUpdate;
    const returnVal = (millis > HOUR_IN_MS * smallerExpiry);
    return returnVal;
  }

  // since the malware ID isn't really a filter list, we need to process it seperately
  if (id === 'malware') {
    // Apply all changes from subData
    for (const property in subData) {
      if (subData[property] !== undefined) {
        this.subscriptions[id][property] = subData[property];
      }
    }

    if (this.subscriptions[id].subscribed) {
      // if forceFetch, set the last update timestamp of the malware to zero, so it's updated now.
      if (forceFetch) {
        this.subscriptions.malware.lastUpdate = 0;
      }

      // load the malware domains
      this.loadMalwareDomains();
    } else {
      if (this.blocking) {
        this.blocking.setMalwareDomains(null);
      }

      // If unsubscribed, remove properties
      delete this.subscriptions[id].text;
      delete this.subscriptions[id].lastUpdate;
      delete this.subscriptions[id].expiresAfterHours;
      delete this.subscriptions[id].lastUpdateFailedAt;
      delete this.subscriptions[id].lastModified;
    }

    this.onSubscriptionChange(subData.subscribed === false);
    return;
  }

  let newId = id;

  // Working with an unknown list: create the list entry
  if (!this.subscriptions[id]) {
    newId = this.customToDefaultId(id);
    if (/^url:.*/.test(newId)) {
      listDidntExistBefore = true;
      this.subscriptions[newId] = {
        userSubmitted: true,
        initialUrl: newId.substr(4),
        url: newId.substr(4),
        title: subData.title,
      };
    }

    subscribeRequiredListToo = true;
  }

  // Subscribing to a well known list should also subscribe to a required list
  if (!this.subscriptions[newId].subscribed && subData.subscribed) {
    subscribeRequiredListToo = true;
  }

  // Apply all changes from subData
  for (const property in subData) {
    if (subData[property] !== undefined) {
      this.subscriptions[newId][property] = subData[property];
    }
  }

  // Check if the required list is a well known list, but only if it is changed
  if (subData.requiresList) {
    // eslint-disable-next-line max-len
    this.subscriptions[newId].requiresList = this.customToDefaultId(this.subscriptions[newId].requiresList);
  }

  if (forceFetch) {
    delete this.subscriptions[newId].lastModified;
  }

  if (this.subscriptions[newId].subscribed) {
    if ((!this.subscriptions[newId].text)
         || outOfDate(this.subscriptions[newId])) {
      this.fetchAndUpdate(newId, listDidntExistBefore);
    }
  } else {
    // If unsubscribed, remove some properties
    delete this.subscriptions[newId].text;
    delete this.subscriptions[newId].lastUpdate;
    delete this.subscriptions[newId].expiresAfterHours;
    delete this.subscriptions[newId].lastUpdateFailedAt;
    delete this.subscriptions[newId].lastModified;
    if (this.subscriptions[newId].deleteMe) {
      // Need to remove the filter list from memory, and permanent storage
      delete this.subscriptions[newId];
      browser.storage.local.remove(keyPrefix + newId);
    }
  }

  // Notify of change.  If we subscribed, we rebuilt above; so we
  // only force a rebuild if we unsubscribed.
  this.onSubscriptionChange(subData.subscribed === false);

  // Subscribe to a required list if nessecary
  if (subscribeRequiredListToo && this.subscriptions[newId]
    && this.subscriptions[newId].requiresList) {
    let requiredList = this.subscriptions[newId].requiresList;
    // If the required list is easylist, but the user is subscribed to easylist_lite
    // change the required list to easylist lite
    if (requiredList === 'easylist'
        && this.subscriptions.easylist_lite.subscribed) {
      requiredList = 'easylist_lite';
    }
    this.changeSubscription(requiredList, { subscribed: true });
  }
};

// Fetch a filter list and parse it
// id:        the id of the list
// isNewList: true when the list is completely new and must succeed or
//            otherwise it'll be deleted.
MyFilters.prototype.fetchAndUpdate = function myFiltersFetchUpdate(id, isNewList) {
  const { url } = this.subscriptions[id];
  const localThis = this;
  function onError() {
    localThis.subscriptions[id].lastUpdateFailedAt = Date.now();
    localThis.onSubscriptionChange();
    if (isNewList) {
      // Delete the list. The user subscribed to an invalid list URL.
      // Delay it a bit, so subscribe.html can first finish it's async call to
      // see if the subscription succeeded or not. Otherwise it assumes it was
      // a well-known subscription and will report a succesfull fetch
      window.setTimeout(() => {
        localThis.changeSubscription(id, { subscribed: false, deleteMe: true });
      }, 500);
    }
  }

  const ajaxRequest = {
    jsonp: false,
    url,
    cache: false,
    headers: {
      Accept: 'text/plain',
      'X-Client-ID': `AdBlock/${STATS.version}`,
      'If-Modified-Since': this.subscriptions[id].lastModified || undefined,
    },
    success(text, status, xhr) {
      // In case the subscription disappeared while we were out
      if (!localThis.subscriptions[id]
          || !localThis.subscriptions[id].subscribed) {
        return;
      }

      // Sometimes text is "". Happens sometimes.  Weird, I know.
      // Every legit list starts with a comment.
      if (status === 'notmodified') {
        log('List not modified ', id, url);
        localThis.updateSubscriptionText(id, localThis.subscriptions[id].text);
        localThis.onSubscriptionChange(true);
      } else if (text
                && (((typeof text === 'string')
                    && text.length !== 0 && Filter.isComment(text.trim()))
                  || (typeof text === 'object'))) {
        log('Fetched ', id, url);
        localThis.updateSubscriptionText(id, text, xhr);
        localThis.onSubscriptionChange(true);
      } else {
        log(`Fetched, but invalid list ${url}`);
        onError();
      }
    },

    error(xhr, textStatus, errorThrown) {
      if (localThis.subscriptions[id]) {
        onError();
      }
      log(`Error fetching ${url}`);
      log(`textStatus ${textStatus}`);
      log(`errorThrown ${errorThrown}`);
    },
  };
  $.ajax(ajaxRequest);
};

// Record that subscription_id is subscribed, was updated now, and has
// the given text.  Requires that this.subscriptions[subscription_id] exists.
// The xhr variable can be used to search the response headers
MyFilters.prototype.updateSubscriptionText = function myFiltersUpdateSubText(id, text, xhr) {
  this.subscriptions[id].lastUpdate = Date.now();
  delete this.subscriptions[id].lastUpdateFailedAt;

  // In case the resource wasn't modified, there is no need to reparse this.
  // xhr isn't send in this case. Do reparse .text, in case we had some update
  // which modified the checks in filternormalizer.js.
  if (xhr) {
    // Store the last time a resource was modified on the server, so we won't re-
    // fetch if it wasn't modified. It is null if the server doesn't support this.
    this.subscriptions[id].lastModified = xhr.getResponseHeader('Last-Modified');

    // Record how many hours until we need to update the subscription text. This
    // can be specified in the file. Defaults to 120.
    this.subscriptions[id].expiresAfterHours = 120;
    const checkLines = text.split('\n', 15); // 15 lines should be enough
    const expiresRegex = /(?:expires:|expires after ) *(\d+) ?(h?)/i;
    const redirectRegex = /(?:redirect:|redirects to ) *(https?:\/\/\S+)/i;
    for (let i = 0; i < checkLines.length; i += 1) {
      if (Filter.isComment(checkLines[i])) {
        let match = checkLines[i].match(redirectRegex);
        if (match && match[1] !== this.subscriptions[id].url) {
          [, this.subscriptions[id].url] = match; // assuming the URL is always correct

          // Force an update.  Even if our refetch below fails we'll have to
          // fetch the new URL in the future until it succeeds.
          this.subscriptions[id].lastUpdate = 0;
        }

        match = checkLines[i].match(expiresRegex);
        if (match && parseInt(match[1], 10)) {
          const hours = parseInt(match[1], 10) * (match[2] === 'h' ? 1 : 24);
          this.subscriptions[id].expiresAfterHours = Math.min(hours, 21 * 24); // 3 week maximum
        }
      }
    }

    // Smear expiry (Issue 7443)
    this.subscriptions[id].expiresAfterHoursHard = this.subscriptions[id].expiresAfterHours * 2;
    const smear = (Math.random() * 0.4) + 0.8;
    this.subscriptions[id].expiresAfterHours *= smear;
  }

  this.subscriptions[id].text = FilterNormalizer.normalizeList(text,
    false, this.subscriptions[id].allowSnippets);

  // The url changed. Simply refetch...
  if (this.subscriptions[id].lastUpdate === 0) {
    this.changeSubscription(id, {}, true);
  }
};

// Checks if subscriptions have to be updated
// Inputs: force? (boolean), true if every filter has to be updated
MyFilters.prototype.checkFilterUpdates = function myFiltersCheckFilterUpdates(force) {
  const key = 'last_subscriptions_check';
  const now = Date.now();
  const forceLocal = force;
  const localThis = this;
  browser.storage.local.get(key, (response) => {
    const delta = now - (response[key] || now);
    const deltaHours = delta / HOUR_IN_MS;
    const saveData = {};
    saveData[key] = now;
    browser.storage.local.set(saveData);
    if (deltaHours > 24) {
      // Extend expiration of subscribed lists (Issue 7443)
      for (const id in localThis.subscriptions) {
        if (localThis.subscriptions[id].subscribed) {
          localThis.subscriptions[id].expiresAfterHours += deltaHours;
        }
      }

      localThis.onSubscriptionChange(); // Store the change
    }

    for (const id in localThis.subscriptions) {
      if (localThis.subscriptions[id].subscribed) {
        localThis.changeSubscription(id, {}, forceLocal);
      }
    }
  });
};

// Checks if a custom id is of a known list
// Inputs: id: the list id to compare
// Returns the id that should be used
MyFilters.prototype.customToDefaultId = function myFiltersCustomToDefaultId(id) {
  const urlOfCustomList = id.substr(4);
  for (const defaultList in this.officialOptions) {
    if (this.officialOptions[defaultList].url === urlOfCustomList) {
      return defaultList;
    }
  }
  return id;
};

// Retreive the list of malware domains from our site.
// and set the response (list of domains) on the blocking
// filter set for processing.
// Retreive the list of malware domains from our site.
// and set the response (list of domains) on the blocking
// filter set for processing.
MyFilters.prototype.loadMalwareDomains = function myFilterLoadMalwareDomains() {
  function outOfDate(subscription) {
    // After a failure, wait at least a day to refetch (overridden below if
    // it has no .text)
    const failedAt = subscription.lastUpdateFailedAt || 0;
    if (Date.now() - failedAt < HOUR_IN_MS * 24) {
      return false;
    }
    const hardStop = subscription.expiresAfterHoursHard || 240;
    const smallerExpiry = Math.min((subscription.expiresAfterHours || 24), hardStop);
    const millis = Date.now() - (subscription.lastUpdate || 0);
    return (millis > HOUR_IN_MS * smallerExpiry);
  }

  if (!this.subscriptions.malware.text
        || !this.getMalwareDomains()
        || outOfDate(this.subscriptions.malware)) {
    const url = `${this.subscriptions.malware.url}?_=${new Date().getTime()}`;

    // Fetch file with malware-known domains
    const xhr = new XMLHttpRequest();
    const localThis = this;
    xhr.onerror = function onErrorFunction() {
      // if the request fail, retry the next time
      localThis.subscriptions.malware.lastUpdateFailedAt = Date.now();
    };

    xhr.onload = function onLoadFunction() {
      if (!xhr.responseText) {
        localThis.subscriptions.malware.lastUpdateFailedAt = Date.now();
        return;
      }

      // Convert the filter list to an array of domains
      const domains = [];
      const lines = xhr.responseText.split('\n');
      for (let i = 0; i < lines.length; i += 1) {
        let line = lines[i];

        // The entries we're interested in, should be of the form: ||domain.org^
        // Ignore all others
        if (line.charAt(0) !== '!'
                && line.charAt(0) !== '['
                && line.charAt(0) === '|'
                && line.charAt(1) === '|') {
          line = line.substr(2); // remove preceeding ||
          line = line.substring(0, line.length - 1); // # remove ending ^
          // convert any non-ascii characters to ascii (punycode)
          line = getUnicodeDomain(line.toLowerCase());
          domains.push(line);
        }
      }

      const malwareObj = {};
      malwareObj.adware = domains;

      // make sure the blocking filter set exists (it may not in Safari 9)
      if (localThis.blocking) {
        localThis.blocking.setMalwareDomains(malwareObj);
      }

      // set the malware object on the 'text' property so it's persisted to storage
      localThis.subscriptions.malware.text = malwareObj;
      localThis.subscriptions.malware.lastUpdate = Date.now();
      localThis.subscriptions.malware.lastModified = Date.now();
      delete localThis.subscriptions.malware.lastUpdateFailedAt;

      // since the Malware file is updated ~1 day
      // expiration is around 24 hours.
      localThis.subscriptions.malware.expiresAfterHours = 24;
      const smear = (Math.random() * 0.4) + 0.8;
      localThis.subscriptions.malware.expiresAfterHours *= smear;
      browser.runtime.sendMessage({ command: 'filters_updated' });
      log(`Fetched ${url}`);
    };

    xhr.open('GET', url);
    xhr.send();
  }
};

// Retreive the list of malware domains from our site.
// and set the response (list of domains) on the blocking
// filter set for processing.
MyFilters.prototype.initializeMalwareDomains = function myFiltersInitMalwareDomains() {
  if (this.subscriptions.malware.text) {
    this.blocking.setMalwareDomains(this.subscriptions.malware.text);
  } else {
    this.loadMalwareDomains();
  }
};

// Get the current list of malware domains
// will return undefined, if the user is not subscribed to the Malware 'filter list'.
MyFilters.prototype.getMalwareDomains = function myFIltersGetMalwareDomains() {
  // make sure the blocking filter set exists (it may not in Safari 9)
  if (this.blocking) {
    return this.blocking.getMalwareDomains();
  }
  return this.subscriptions.malware.text;
};

// If the user wasn't subscribed to any lists, subscribe to
// EasyList, AdBlock custom and (if any) a localized subscription
// Inputs: none.
// Returns an object containing the subscribed lists
MyFilters.prototype.loadDefaultSubscriptions = function myFiltersLoadDefaultSubs() {
  const result = {};

  // Returns the ID of the list appropriate for the user's locale, or ''
  function listIdForThisLocale() {
    const language = determineUserLanguage();

    if (language === 'pt') {
      const country = navigator.language.match(/([A-Z]+$)/g);
      if (country && country[0] === 'BR') {
        return 'brazilian_portuguese';
      }
    }

    switch (language) {
      case 'ar': return 'easylist_plus_arabic';
      case 'bg': return 'easylist_plus_bulgarian';
      case 'cs': return 'czech';
      case 'cu': return 'easylist_plus_bulgarian';
      case 'da': return 'danish';
      case 'de': return 'easylist_plus_german';
      case 'el': return 'easylist_plus_greek';
      case 'es': return 'easylist_plus_spanish';
      case 'et': return 'easylist_plus_estonian';
      case 'fr': return 'easylist_plus_french';
      case 'he': return 'israeli';
      case 'hu': return 'hungarian';
      case 'is': return 'icelandic';
      case 'it': return 'italian';
      case 'id': return 'easylist_plus_indonesian';
      case 'ja': return 'japanese';
      case 'ko': return 'easylist_plun_korean';
      case 'lt': return 'easylist_plus_lithuania';
      case 'lv': return 'latvian';
      case 'nl': return 'dutch';
      case 'pl': return 'easylist_plus_polish';
      case 'ro': return 'easylist_plus_romanian';
      case 'ru': return 'russian';
      case 'sk': return 'czech';
      case 'sv': return 'swedish';
      case 'tr': return 'turkish';
      case 'uk': return 'russian';
      case 'zh': return 'chinese';
      case 'nb': return 'norwegian';
      case 'vi': return 'easylist_plus_vietnamese';
      default: return '';
    }
  }

  // Update will be done immediately after this function returns
  result.adblock_custom = { subscribed: true };
  if (Fennec) {
    result.easylist_lite = { subscribed: true };
  } else {
    result.easylist = { subscribed: true };
  }

  result.acceptable_ads = { subscribed: true };
  result.bitcoin_mining_protection = { subscribed: true };
  result.anticircumvent = { subscribed: true };
  const listForLang = listIdForThisLocale();
  if (listForLang) {
    result[listForLang] = { subscribed: true };
  }
  return result;
};

// Used to create the list of default subscriptions
// Called when MyFilters is created.
// Returns: that list
MyFilters.prototype.makeSubscriptionOptions = function myFiltersMakeSubOptions() {
  // When modifying a list, IDs mustn't change!
  return {
    adblock_custom: { // AdBlock custom filters
      url: 'https://cdn.adblockcdn.com/filters/adblock_custom.txt',
    },
    easylist: { // EasyList
      url: 'https://easylist-downloads.adblockplus.org/easylist.txt',
    },
    easylist_lite: { // EasyList lite
      url: 'https://cdn.adblockcdn.com/filters/easylist_lite.txt',
    },
    easylist_plus_bulgarian: { // Additional Bulgarian filters
      url: 'https://easylist-downloads.adblockplus.org/bulgarian_list+easylist.txt',
      requiresList: 'easylist',
    },
    dutch: { // Additional Dutch filters
      url: 'https://easylist-downloads.adblockplus.org/easylistdutch.txt',
      requiresList: 'easylist',
    },
    easylist_plus_french: { // Additional French filters
      url: 'https://easylist-downloads.adblockplus.org/liste_fr.txt',
      requiresList: 'easylist',
    },
    easylist_plus_german: { // Additional German filters
      url: 'https://easylist-downloads.adblockplus.org/easylistgermany.txt',
      requiresList: 'easylist',
    },
    easylist_plus_greek: { // Additional Greek filters
      url: 'https://www.void.gr/kargig/void-gr-filters.txt',
      requiresList: 'easylist',
    },
    easylist_plus_indonesian: { // Additional Indonesian filters
      url: 'https://easylist-downloads.adblockplus.org/abpindo+easylist.txt',
    },
    easylist_plus_polish: { // Additional Polish filters
      url: 'https://raw.githubusercontent.com/MajkiIT/polish-ads-filter/master/polish-adblock-filters/adblock.txt',
      requiresList: 'easylist',
    },
    easylist_plus_romanian: { // Additional Romanian filters
      url: 'https://www.zoso.ro/pages/rolist.txt',
      requiresList: 'easylist',
    },
    russian: { // Additional Russian filters
      url: 'https://easylist-downloads.adblockplus.org/advblock.txt',
      requiresList: 'easylist',
    },
    chinese: { // Additional Chinese filters
      url: 'https://easylist-downloads.adblockplus.org/easylistchina.txt',
      requiresList: 'easylist',
    },
    czech: { // Additional Czech and Slovak filters
      url: 'https://raw.github.com/tomasko126/easylistczechandslovak/master/filters.txt',
      requiresList: 'easylist',
    },
    danish: { // Danish filters
      url: 'https://adblock.schack.dk/block.txt',
    },
    hungarian: { // Hungarian filters
      url: 'https://raw.githubusercontent.com/szpeter80/hufilter/master/hufilter.txt',
    },
    israeli: { // Israeli filters
      url: 'https://easylist-downloads.adblockplus.org/israellist+easylist.txt',
    },
    italian: { // Italian filters
      url: 'https://easylist-downloads.adblockplus.org/easylistitaly.txt',
      requiresList: 'easylist',
    },
    japanese: { // Japanese filters
      url: 'https://bit.ly/11QrCfx',
    },
    easylist_plun_korean: { // Korean filters
      url: 'https://raw.githubusercontent.com/gfmaster/adblock-korea-contrib/master/filter.txt',
      requiresList: 'easylist',
    },
    latvian: { // Latvian filters
      url: 'https://notabug.org/latvian-list/adblock-latvian/raw/master/lists/latvian-list.txt',
    },
    norwegian: { // Norwegian filters
      url: 'https://raw.githubusercontent.com/DandelionSprout/adfilt/master/NorwegianList.txt',
    },
    swedish: { // Swedish filters
      url: 'https://raw.githubusercontent.com/lassekongo83/Frellwits-filter-lists/master/Frellwits-Swedish-Filter.txt',
    },
    turkish: { // Turkish filters
      url: 'https://filters.adtidy.org/extension/chromium/filters/13.txt',
    },
    easyprivacy: { // EasyPrivacy
      url: 'https://easylist-downloads.adblockplus.org/easyprivacy.txt',
    },
    antisocial: { // Antisocial
      url: 'https://easylist-downloads.adblockplus.org/fanboy-social.txt',
    },
    malware: { // Malware protection
      url: 'https://easylist-downloads.adblockplus.org/malwaredomains_full.txt',
    },
    annoyances: { // Fanboy's Annoyances
      url: 'https://easylist-downloads.adblockplus.org/fanboy-annoyance.txt',
    },
    warning_removal: { // AdBlock warning removal
      url: 'https://easylist-downloads.adblockplus.org/antiadblockfilters.txt',
    },
    acceptable_ads: { // Acceptable Ads
      url: 'https://easylist-downloads.adblockplus.org/exceptionrules.txt',
    },
    easylist_plus_estonian: { // Estonian filters
      url: 'http://gurud.ee/ab.txt',
      requiresList: 'easylist',
    },
    easylist_plus_lithuania: { // Lithuania filters
      url: 'http://margevicius.lt/easylistlithuania.txt',
      requiresList: 'easylist',
    },
    easylist_plus_arabic: { // Arabic filters
      url: 'https://easylist-downloads.adblockplus.org/Liste_AR.txt',
      requiresList: 'easylist',
    },
    icelandic: { // Icelandic filters
      url: 'https://adblock.gardar.net/is.abp.txt',
    },
    easylist_plus_spanish: { // Spanish filters
      url: 'https://easylist-downloads.adblockplus.org/easylistspanish.txt',
      requiresList: 'easylist',
    },
    bitcoin_mining_protection: {
      url: 'https://raw.githubusercontent.com/hoshsadiq/adblock-nocoin-list/master/nocoin.txt',
    },
    easylist_plus_vietnamese: { // Vietnamese filters
      url: 'https://easylist-downloads.adblockplus.org/abpvn+easylist.txt',
    },
    brazilian_portuguese: { // Brazilian Portuguese filters
      url: 'http://easylistbrasil.freevar.com/filtro/easylistbrasil.php',
    },
    anticircumvent: { // Anti-cirucumvention
      url: 'https://easylist-downloads.adblockplus.org/abp-filters-anti-cv.txt',
      allowSnippets: true,
    },
  };
};

/* subscription properties:
url (string): url of subscription
initialUrl (string): the hardcoded url. Same as .url except when redirected
userSubmitted (bool): submitted by the user or not
requiresList (string): id of a list required for this list
subscribed (bool): if you are subscribed to the list or not
lastUpdate (date): time of the last succesfull update
lastModified (string): time of the last change on the server
lastUpdateFailedAt (date): if set, when the last update attempt failed
text (string): the filters of the subscription
expiresAfterHours (int): the time after which the subscription expires
expiresAfterHoursHard (int): we must redownload subscription after this delay
deleteMe (bool): if the subscription has to be deleted
allowSnippets (bool): if the subscription can have snippet rules
*/

// Save the current subscription information to storage
MyFilters.prototype.saveSubscriptions = function myFiltersSaveSubs() {
  const result = {};
  for (const id in this.subscriptions) {
    result[keyPrefix + id] = {};
    for (const attr in this.subscriptions[id]) {
      result[keyPrefix + id][attr] = this.subscriptions[id][attr];
    }
  }

  browser.storage.local.set(result, () => {
    if (browser.runtime.lastError) {
      log('write failure', browser.runtime.lastError);
    }
  });
};

MyFilters.prototype.getSubscriptions = function myFiltersGetSubs(callback) {
  const localThis = this;
  browser.storage.local.get(undefined, (readResponse) => {
    if (browser.runtime.lastError) {
      log('read failure', browser.runtime.lastError);
    }

    const readResult = {};
    for (const fullID in readResponse) {
      if (fullID.startsWith(keyPrefix)) {
        const id = fullID.substring(keyPrefixLength);
        readResult[id] = {};
        for (const attr in readResponse[fullID]) {
          readResult[id][attr] = readResponse[fullID][attr];
        }
      }
    }

    localThis.newUser = (Object.keys(readResult).length === 0);
    if (typeof callback === 'function') {
      callback(readResult);
    }
  });
};
